"""Implements section for isort."""
from dataclasses import dataclass
from typing import ClassVar, Optional

from pyvelocity.configurations.files.sections import ConfigurationFileParameter, Section


@dataclass
class Isort(Section):
    NAME: ClassVar[str] = "isort"
    LIST_PARAMETER_NAME: ClassVar[list[str]] = ["line_length"]
    line_length: ConfigurationFileParameter[Optional[int]]



"""Implements pyvelocity configurations."""
from typing import Optional

from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.files.sections import (
    ConfigurationFileParameter,
    WhereToolDefault,
    is_not_none_value,
    pyvelocity,
)
from pyvelocity.configurations.tools import Tool


class Pyvelocity(Tool):
    """pyvelocity configurations."""

    NAME = "pyvelocity"

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.filter: ConfigurationFileParameter[list[str]] = ConfigurationFileParameter(
            WhereToolDefault(self), ConfigurationFileParameter.NAME_TOOL_DEFAULT, []
        )
        if configuration_files.py_project_toml:
            self.overwrite(configuration_files.py_project_toml.pyvelocity)

    def overwrite(self, section_pyvelocity: Optional[pyvelocity.Pyvelocity]):
        if section_pyvelocity:
            if is_not_none_value(section_pyvelocity.filter):
                self.filter = section_pyvelocity.filter



"""Implements configuration file sections."""
import sys
from abc import ABC
from dataclasses import dataclass
from typing import ClassVar, Generic, Optional, TypeVar

from pyvelocity.configurations.files import ConfigurationFile
from pyvelocity.configurations.tools import Tool

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard  # pragma: no cover


@dataclass
class Section(ABC):
    NAME: ClassVar[str]
    LIST_PARAMETER_NAME: ClassVar[list[str]]
    configuration_file: ConfigurationFile


@dataclass
class Where(ABC):
    pass


@dataclass
class WhereFile(Where):
    """Information of where about file."""

    file: ConfigurationFile
    section: type[Section]
    node: Optional[str]

    def __str__(self) -> str:
        if self.node:
            return f"{self.file.name} {self.node}.{self.section.NAME}"
        return f"{self.file.name} {self.section.NAME}"


@dataclass
class WhereToolDefault(Where):
    tool: Tool

    def __str__(self) -> str:
        return self.tool.NAME


T = TypeVar("T")


@dataclass
class ConfigurationFileParameter(Generic[T]):
    """The parameter of configuration file."""

    NAME_TOOL_DEFAULT = "tool default"
    where: Where
    name: str
    value: T

    @property
    def full_name(self) -> str:
        return f"{self.where} {self.name}"


def is_not_none_value(instance: ConfigurationFileParameter[Optional[T]]) -> TypeGuard[ConfigurationFileParameter[T]]:
    """This function can't be implemented as instance method.

    Since TypeGuard does not narrow types of self or cls implicit arguments.

    see:
    - Type narrowing - mypy 0.980+dev.c2949e969ffcba848595d0851a9193fe1bc3e0a1.dirty documentation
      https://mypy.readthedocs.io/en/latest/type_narrowing.html#typeguards-as-methods
    """
    return instance.value is not None



"""Implements Pylint configurations."""
from typing import Optional

from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.files.sections import (
    ConfigurationFileParameter,
    WhereToolDefault,
    is_not_none_value,
    pylint,
)
from pyvelocity.configurations.tools import Tool


class Format(Tool):
    """Pylint configurations."""

    NAME = "Pylint"

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.max_line_length = ConfigurationFileParameter(
            WhereToolDefault(self), ConfigurationFileParameter.NAME_TOOL_DEFAULT, 100
        )
        if configuration_files.setup_cfg and configuration_files.setup_cfg.pylint:
            self.overwrite(configuration_files.setup_cfg.pylint.format)
        if configuration_files.py_project_toml and configuration_files.py_project_toml.pylint:
            self.overwrite(configuration_files.py_project_toml.pylint.format)

    def overwrite(self, pylint_format: Optional[pylint.Format]) -> None:
        if pylint_format:
            if is_not_none_value(pylint_format.max_line_length):
                self.max_line_length = pylint_format.max_line_length


# Reason: Aggregation of Pylint sections. pylint: disable=too-few-public-methods
class Pylint:
    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.format = Format(configuration_files)



"""Implements sections for Pylint."""
from configparser import ConfigParser
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Optional

from pyvelocity.configurations.files.sections import ConfigurationFileParameter, Section
from pyvelocity.configurations.files.sections.factory import PyProjectTomlSectionFactory, SetupCfgSectionFactory

if TYPE_CHECKING:  # pragma: no cover
    from pyvelocity.configurations.files.py_project_toml import PyProjectToml
    from pyvelocity.configurations.files.setup_cfg import SetupCfg


@dataclass
class Format(Section):
    NAME: ClassVar[str] = "format"
    LIST_PARAMETER_NAME: ClassVar[list[str]] = ["max-line-length"]
    max_line_length: ConfigurationFileParameter[Optional[int]]


# Reason: Aggregation class. pylint: disable=too-many-instance-attributes
@dataclass
class Pylint:
    """Node of pylint."""

    NAME: ClassVar[str] = "pylint"
    main: None
    basic: None
    classes: None
    design: None
    exceptions: None
    format: Optional[Format]
    imports: None
    logging: None
    method_args: None
    miscellaneous: None
    refactoring: None
    similarities: None
    spelling: None
    string: None
    typecheck: None
    variables: None
    broad_try_clause: None
    code_style: None
    deprecated_builtins: None
    parameter_documentation: None
    typing: None


class PyProjectTomlPylintFactory:
    """Factory for pylint node."""

    @staticmethod
    def create(
        py_project_toml: "PyProjectToml", node: str, tool: dict[str, Optional[dict[str, Any]]]
    ) -> Optional[Pylint]:
        """Creates node of Pylint."""
        config = tool.get(Pylint.NAME)
        if not config:
            return None
        return Pylint(
            None,
            None,
            None,
            None,
            None,
            PyProjectTomlSectionFactory.create(py_project_toml, f"{node}.{Pylint.NAME}", Format, config),
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )


class SetupCfgPylintFactory:
    """Factory for pylint node."""

    @staticmethod
    def create(setup_cfg: "SetupCfg", config_parser: ConfigParser) -> Optional[Pylint]:
        """Creates node of Pylint."""
        pylint = Pylint(
            None,
            None,
            None,
            None,
            None,
            SetupCfgSectionFactory.create(setup_cfg, Pylint.NAME, Format, config_parser),
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )
        if all(v is None for v in pylint.__dict__.values()):
            return None
        return pylint



"""Implements aggregation of configuration files."""
from pathlib import Path

from pyvelocity.configurations.files.py_project_toml import WHERE_PY_PROJECT_TOML, PyProjectToml
from pyvelocity.configurations.files.setup_cfg import WHERE_SETUP_CFG, SetupCfg


# Reason: Aggregation class. pylint: disable=too-few-public-methods
class ConfigurationFiles:
    def __init__(self) -> None:
        path_py_project_toml = Path(WHERE_PY_PROJECT_TOML)
        self.py_project_toml = PyProjectToml(path_py_project_toml) if path_py_project_toml.exists() else None
        path_setup_cfg = Path(WHERE_SETUP_CFG)
        self.setup_cfg = SetupCfg(path_setup_cfg) if path_setup_cfg.exists() else None



"""Implements pyproject.toml."""
from pathlib import Path

import tomli

from pyvelocity.configurations.files import ConfigurationFile
from pyvelocity.configurations.files.sections.black import Black
from pyvelocity.configurations.files.sections.docformatter import Docformatter
from pyvelocity.configurations.files.sections.factory import PyProjectTomlSectionFactory
from pyvelocity.configurations.files.sections.isort import Isort
from pyvelocity.configurations.files.sections.pylint import PyProjectTomlPylintFactory
from pyvelocity.configurations.files.sections.pyvelocity import Pyvelocity

WHERE_PY_PROJECT_TOML = "pyproject.toml"


class PyProjectToml(ConfigurationFile):
    """pyproject.toml."""

    def __init__(self, path_py_project_toml: Path) -> None:
        super().__init__()
        parsed_toml = tomli.loads(path_py_project_toml.read_text())
        node_tool = "tool"
        tool = parsed_toml[node_tool]
        self.black = PyProjectTomlSectionFactory.create(self, node_tool, Black, tool)
        self.docformatter = PyProjectTomlSectionFactory.create(self, node_tool, Docformatter, tool)
        self.isort = PyProjectTomlSectionFactory.create(self, node_tool, Isort, tool)
        self.pylint = PyProjectTomlPylintFactory.create(self, node_tool, tool)
        self.pyvelocity = PyProjectTomlSectionFactory.create(self, node_tool, Pyvelocity, tool)

    @property
    def name(self) -> str:
        return WHERE_PY_PROJECT_TOML



"""Implements Flake8 configurations."""
from typing import Optional

from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.files.sections import (
    ConfigurationFileParameter,
    WhereToolDefault,
    flake8,
    is_not_none_value,
)
from pyvelocity.configurations.tools import Tool


class Flake8(Tool):
    """Flake8 configurations."""

    NAME = "Flake8"

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.max_line_length = ConfigurationFileParameter(
            WhereToolDefault(self), ConfigurationFileParameter.NAME_TOOL_DEFAULT, 79
        )
        if configuration_files.setup_cfg:
            self.overwrite(configuration_files.setup_cfg.flake8)

    def overwrite(self, section_flake8: Optional[flake8.Flake8]):
        if section_flake8:
            if is_not_none_value(section_flake8.max_line_length):
                self.max_line_length = section_flake8.max_line_length



"""Implements configuration files."""
from abc import ABC, abstractmethod


class ConfigurationFile(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError  # pragma: no cover



"""Implements docformatter configurations.

see:
  - Welcome to docformatter! — docformatter documentation
    https://docformatter.readthedocs.io/en/latest/
"""
from typing import Optional

from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.files.sections import (
    ConfigurationFileParameter,
    WhereToolDefault,
    docformatter,
    is_not_none_value,
)
from pyvelocity.configurations.tools import Tool


class Docformatter(Tool):
    """docformatter configurations."""

    NAME = "docformatter"

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        where_tool_default = WhereToolDefault(self)
        self.wrap_descriptions = ConfigurationFileParameter(
            where_tool_default, ConfigurationFileParameter.NAME_TOOL_DEFAULT, 72
        )
        self.wrap_summaries = ConfigurationFileParameter(
            where_tool_default, ConfigurationFileParameter.NAME_TOOL_DEFAULT, 79
        )
        if configuration_files.setup_cfg:
            self.overwrite(configuration_files.setup_cfg.docformatter)
        if configuration_files.py_project_toml:
            self.overwrite(configuration_files.py_project_toml.docformatter)

    def overwrite(self, section_docformatter: Optional[docformatter.Docformatter]):
        if section_docformatter:
            if is_not_none_value(section_docformatter.wrap_descriptions):
                self.wrap_descriptions = section_docformatter.wrap_descriptions
            if is_not_none_value(section_docformatter.wrap_summaries):
                self.wrap_summaries = section_docformatter.wrap_summaries



"""Implements isort."""
from typing import Optional

from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.files.sections import (
    ConfigurationFileParameter,
    WhereToolDefault,
    is_not_none_value,
    isort,
)
from pyvelocity.configurations.tools import Tool


class Isort(Tool):
    """isort configurations."""

    NAME = "isort"

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.line_length = ConfigurationFileParameter(
            WhereToolDefault(self), ConfigurationFileParameter.NAME_TOOL_DEFAULT, 79
        )
        if configuration_files.setup_cfg:
            self.overwrite(configuration_files.setup_cfg.isort)
        if configuration_files.py_project_toml:
            self.overwrite(configuration_files.py_project_toml.isort)

    def overwrite(self, section_isort: Optional[isort.Isort]):
        if section_isort:
            if is_not_none_value(section_isort.line_length):
                self.line_length = section_isort.line_length



"""implements aggregation of tools."""
from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.tools.black import Black
from pyvelocity.configurations.tools.docformatter import Docformatter
from pyvelocity.configurations.tools.flake8 import Flake8
from pyvelocity.configurations.tools.isort import Isort
from pyvelocity.configurations.tools.pylint import Pylint
from pyvelocity.configurations.tools.pyvelocity import Pyvelocity


# Reason: Aggregation class. pylint: disable=too-few-public-methods
class Configurations:
    """Aggregation of tools."""

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.black = Black(configuration_files)
        self.docformatter = Docformatter(configuration_files)
        self.flake8 = Flake8(configuration_files)
        self.isort = Isort(configuration_files)
        self.pylint = Pylint(configuration_files)
        self.pyvelocity = Pyvelocity(configuration_files)



"""Implements section for pyvelocity."""
from dataclasses import dataclass
from typing import ClassVar, Optional

from pyvelocity.configurations.files.sections import ConfigurationFileParameter, Section


@dataclass
class Pyvelocity(Section):
    NAME: ClassVar[str] = "pyvelocity"
    LIST_PARAMETER_NAME: ClassVar[list[str]] = ["filter"]
    filter: ConfigurationFileParameter[Optional[list[str]]]



"""Implements using pyproject.toml check."""
from pyvelocity.checks import Check, Result


class UsingPyProjectToml(Check):
    ID = "using-py-project-toml"

    def execute(self) -> Result:
        is_ok = self.configuration_files.py_project_toml is not None
        message = "" if is_ok else ("It's recommended to use pyproject.toml to gather settings for project.")
        return Result(self.ID, is_ok, message)



"""Implements section for docformatter."""
from dataclasses import dataclass
from typing import ClassVar, Optional

from pyvelocity.configurations.files.sections import ConfigurationFileParameter, Section


@dataclass
class Docformatter(Section):
    """Section of docformatter."""

    NAME: ClassVar[str] = "docformatter"
    LIST_PARAMETER_NAME: ClassVar[list[str]] = [
        "wrap-descriptions",
        "wrap-summaries",
    ]
    wrap_descriptions: ConfigurationFileParameter[Optional[int]]
    wrap_summaries: ConfigurationFileParameter[Optional[int]]



"""Console script for pyvelocity."""

import click
from click import ClickException

from pyvelocity.checks.aggregation import Checks, Results
from pyvelocity.configurations.aggregation import Configurations
from pyvelocity.configurations.files.aggregation import ConfigurationFiles


@click.command()
def main() -> None:
    """Console script for pyvelocity."""
    configuration_files = ConfigurationFiles()
    configurations = Configurations(configuration_files)
    results = Results(list(Checks(configuration_files, configurations).execute()))
    if results.message:
        click.echo(results.message)
    if not results.is_ok:
        exception = ClickException("Looks there are some of improvements.")
        exception.exit_code = 3
        raise exception
    click.echo("Looks high velocity! ⚡️ 🚄 ✨")



"""Implements section factory."""
from typing import TYPE_CHECKING, Any, MutableMapping, Optional, TypeVar

from pyvelocity.configurations.files import ConfigurationFile
from pyvelocity.configurations.files.sections import ConfigurationFileParameter, Section, WhereFile

if TYPE_CHECKING:  # pragma: no cover
    from pyvelocity.configurations.files.py_project_toml import PyProjectToml
    from pyvelocity.configurations.files.setup_cfg import SetupCfg

TypeVarSection = TypeVar("TypeVarSection", bound=Section)


class SectionFactory:
    """Factory for Section instance."""

    def __init__(
        self,
        configuration_file: ConfigurationFile,
        node: Optional[str],
        class_section: type[TypeVarSection],
        config: dict[str, Any],
    ) -> None:
        self.configuration_file = configuration_file
        self.node = node
        self.class_section = class_section
        self.config = config

    def create_section(self) -> TypeVarSection:
        configurations = [
            self.create_configuration_parameter(parameter_name)
            for parameter_name in self.class_section.LIST_PARAMETER_NAME
        ]
        return self.class_section(self.configuration_file, *configurations)

    def create_configuration_parameter(self, parameter_name) -> ConfigurationFileParameter:
        return ConfigurationFileParameter(
            WhereFile(self.configuration_file, self.class_section, self.node),
            parameter_name,
            self.config.get(parameter_name),
        )


class PyProjectTomlSectionFactory:
    """Factory for Section instance for PyProjectToml."""

    @classmethod
    def create(
        cls,
        py_project_toml: "PyProjectToml",
        node: str,
        class_configuration: type[TypeVarSection],
        tool: dict[str, Optional[dict[str, Any]]],
    ) -> Optional[TypeVarSection]:
        """Creates Section instance for PyProjectToml."""
        config = tool.get(class_configuration.NAME)
        if not config:
            return None
        return SectionFactory(py_project_toml, node, class_configuration, config).create_section()


class SetupCfgSectionFactory:
    """Factory for Section instance for SetupCfg."""

    @classmethod
    def create(
        cls,
        setup_cfg: "SetupCfg",
        node: Optional[str],
        class_configuration: type[TypeVarSection],
        config_parser: MutableMapping,
    ) -> Optional[TypeVarSection]:
        """Creates Section instance for SetupCfg."""
        try:
            config = dict(config_parser[f"{node}.{class_configuration.NAME}" if node else class_configuration.NAME])
        except KeyError:
            return None
        return SectionFactory(setup_cfg, node, class_configuration, config).create_section()



"""Implements line length check."""
from collections import Counter
from typing import Any, Optional

from pyvelocity.checks import Check, Result
from pyvelocity.configurations.files.sections import ConfigurationFileParameter


class LineLength(Check):
    """Check about line length."""

    ID = "line-length"

    def execute(self) -> Result:
        target_configurations: list[ConfigurationFileParameter] = []
        target_configurations = [
            self.configurations.docformatter.wrap_descriptions,
            self.configurations.docformatter.wrap_summaries,
            self.configurations.isort.line_length,
            self.configurations.black.line_length,
            self.configurations.flake8.max_line_length,
            self.configurations.pylint.format.max_line_length,
        ]
        counter = Counter([target_configuration.value for target_configuration in target_configurations])
        most_common = counter.most_common()[0]
        is_ok = most_common[1] == len(target_configurations)
        message = "" if is_ok else (self.build_message(most_common, target_configurations))
        return Result(self.ID, is_ok, message)

    @staticmethod
    def build_message(
        most_common: tuple[Optional[Any], int], target_configurations: list[ConfigurationFileParameter]
    ) -> str:
        """Builds message."""
        error_messages = [
            f"{target_configuration.full_name} = {str(target_configuration.value)}"
            for target_configuration in target_configurations
            if target_configuration.value != most_common[0]
        ]
        return "Line length are not consistent.\n\t" f"Most common = {most_common[0]}\n\t" + "\n\t".join(error_messages)



"""Implements section for Flake8."""
from dataclasses import dataclass
from typing import ClassVar, Optional

from pyvelocity.configurations.files.sections import ConfigurationFileParameter, Section


@dataclass
class Flake8(Section):
    """Section of flake8."""

    NAME: ClassVar[str] = "flake8"
    LIST_PARAMETER_NAME: ClassVar[list[str]] = ["max-line-length"]
    _max_line_length: ConfigurationFileParameter[Optional[str]]

    @property
    def max_line_length(self) -> ConfigurationFileParameter[Optional[int]]:
        return ConfigurationFileParameter(
            self._max_line_length.where,
            self._max_line_length.name,
            None if self._max_line_length.value is None else int(self._max_line_length.value),
        )



"""Implements setup.cfg."""
from configparser import ConfigParser
from pathlib import Path

from pyvelocity.configurations.files import ConfigurationFile
from pyvelocity.configurations.files.sections.docformatter import Docformatter
from pyvelocity.configurations.files.sections.factory import SetupCfgSectionFactory
from pyvelocity.configurations.files.sections.flake8 import Flake8
from pyvelocity.configurations.files.sections.isort import Isort
from pyvelocity.configurations.files.sections.pylint import SetupCfgPylintFactory

WHERE_SETUP_CFG = "setup.cfg"


class SetupCfg(ConfigurationFile):
    """setup.cfg."""

    def __init__(self, path_setup_cfg: Path) -> None:
        super().__init__()
        config_parser = ConfigParser()
        config_parser.read(path_setup_cfg)
        self.docformatter = SetupCfgSectionFactory.create(self, None, Docformatter, config_parser)
        self.flake8 = SetupCfgSectionFactory.create(self, None, Flake8, config_parser)
        self.isort = SetupCfgSectionFactory.create(self, None, Isort, config_parser)
        self.pylint = SetupCfgPylintFactory.create(self, config_parser)

    @property
    def name(self) -> str:
        return WHERE_SETUP_CFG



"""Implements aggregation of checks."""
from typing import Generator

from pyvelocity.checks import Result
from pyvelocity.checks.line_length import LineLength
from pyvelocity.checks.using_py_project_toml import UsingPyProjectToml
from pyvelocity.configurations.aggregation import Configurations
from pyvelocity.configurations.files.aggregation import ConfigurationFiles


class Results:
    """Aggregation of Result."""

    def __init__(self, results: list[Result]) -> None:
        self.results = results

    @property
    def message(self) -> str:
        return "\n".join([result.message for result in self.results if result.message])

    @property
    def is_ok(self) -> bool:
        return all(result.is_ok for result in self.results)


class Checks:
    """Aggregation of Check."""

    def __init__(self, configuration_files: ConfigurationFiles, configurations: Configurations) -> None:
        checks = [UsingPyProjectToml, LineLength]
        self.checks = (
            check(configuration_files, configurations)
            for check in checks
            if check.ID not in configurations.pyvelocity.filter.value
        )

    def execute(self) -> Generator[Result, None, None]:
        return (check.execute() for check in self.checks)



"""Implements checks."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import ClassVar

from pyvelocity.configurations.aggregation import Configurations
from pyvelocity.configurations.files.aggregation import ConfigurationFiles


@dataclass
class Result:
    # Reason: No problem. pylint: disable=invalid-name
    id: str
    is_ok: bool
    message: str


class Check(ABC):
    """Abstract check class."""

    ID: ClassVar[str]

    def __init__(self, configuration_files: ConfigurationFiles, configurations: Configurations) -> None:
        self.configuration_files = configuration_files
        self.configurations = configurations

    @abstractmethod
    def execute(self) -> Result:
        raise NotImplementedError  # pragma: no cover



"""Implements Black configurations."""
from typing import Optional

from pyvelocity.configurations.files.aggregation import ConfigurationFiles
from pyvelocity.configurations.files.sections import (
    ConfigurationFileParameter,
    WhereToolDefault,
    black,
    is_not_none_value,
)
from pyvelocity.configurations.tools import Tool


class Black(Tool):
    """Black configurations."""

    NAME = "Black"

    def __init__(self, configuration_files: ConfigurationFiles) -> None:
        self.line_length = ConfigurationFileParameter(
            WhereToolDefault(self), ConfigurationFileParameter.NAME_TOOL_DEFAULT, 88
        )
        if configuration_files.py_project_toml:
            self.overwrite(configuration_files.py_project_toml.black)

    def overwrite(self, section_black: Optional[black.Black]):
        if section_black:
            if is_not_none_value(section_black.line_length):
                self.line_length = section_black.line_length
